#*******************************************************************************
# Licensed Materials - Property of IBM
#
# 5737-M65
#
# (C) Copyright IBM Corp. 2020. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#*******************************************************************************

#!/bin/bash

#---------------------------------------
# Check NFS 
#---------------------------------------

check_nfs () {

	nfs_util_cnt=`which nfsstat 2>/dev/null | wc -l`

	if [ $nfs_util_cnt -eq 0 ] 
	then
 		echo "[ERROR] nfs-utils is not installed --- failed"
	else
		echo "[INFO] nfs-utils check         --- ok"
	fi
}

#---------------------------------------
# Check Firewall
#---------------------------------------

function ip_format () {
	
	if [[ "$1" =~ ^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}$ ]]
	then
		echo ipv4 
	else
		if [[ $1 =~ .*:.* ]] 
		then
			echo ipv6
		fi
	fi
}

function hex2bin () {

        local hexa=`echo $1|tr [a-f] [A-F]`
        #echo "ibase=16;$hexa_value"|bc
	binary_value=`printf "%04d%b\n" $(bc <<< "ibase=16; obase=2; $hexa")`
        echo $binary_value
}


function hex2dec () {

        local hexa_value=`echo $1|tr [a-f] [A-F]`
        echo "ibase=16;$hexa_value"|bc
}

function bin2dec () {

        echo "ibase=2;obase=A;$1"|bc
}


function getFullBit() {
	
	local length=$1     # Length is 32 bit [IPv4] or 128 bits [IPv6]
	local fullbit=`echo "2^$length"|bc`

	if [ $length == 128 ]
	then
		fullbit=`expr $fullbit - 1 `   # 
		fullbit=`expr $fullbit \\* -1` # Negative value in order to get the number of host in IPv6 use in Bitwise
	else
		fullbit=`expr $fullbit - 1 `   
	fi
	echo $fullbit
}

function getV4Address () {

	ipv4=$1

	# This function is to convert IPv4 to Decimal

	ip_dec=0
	for octet in `echo $ipv4 |tr "\."  " "`
	do
       		ip_dec=$(($ip_dec<<8 | $octet))
	done
	echo $ip_dec
}

function getV6Address() {

	# This function is to 
	#  a) Convert IPv6 Hexa to Decimal Value
	#  b) Get the IPv6 Prefix Block in decimal value if pass in with length (ie CIDR Length )
      
	#local ipv6=$1
	local ip=`echo $1 | tr [A-F] [a-f]` 	
	local ipv6=$(expand_ipv6 $ip)	
	local length=$2

	local ipv6_bin=""
	local ipv6_dec=""

	IP=`echo $ipv6 | sed 's/://g'`

	for ((i=1; i<=32;i++))  # 32 is the total hexa of IPv6
	do
		hex=$(expr substr "$IP" $i 1)
		ipv6_bin+=$(hex2bin $hex)
	done
	
	# If Prefix Length is pass in , get IP binary base on Prefix Length
	# This is to get the Start Range of IPv6
	if [ ! -z "$length" ]
	then
		ipv6_bin=`echo ${ipv6_bin:0:$length} | sed -e :a -e 's/^.\{1,127\}$/&0/;ta'`  # Add Right Padding Zero 
	fi

	ipv6_dec=$(bin2dec $ipv6_bin)
	
	echo $ipv6_dec
	
}

function expand_ipv6() {
    ip=$1

    # Example 2001:db8:a0b:12f0::1238 
    # Become  2001:0db8:a0b:12f0:0000:0000:0000:1238

    # prepend 0 if we start with :
    echo $ip | grep -qs "^:" && ip="0${ip}"

    # expand ::
    if echo $ip | grep -qs "::"; then
        colons=$(echo $ip | sed 's/[^:]//g')
        missing=$(echo ":::::::::" | sed "s/$colons//")
        expanded=$(echo $missing | sed 's/:/:0/g')
        ip=$(echo $ip | sed "s/::/$expanded/")
    fi

    blocks=$(echo $ip | grep -o "[0-9a-f]\+")
    set $blocks

    printf "%04x:%04x:%04x:%04x:%04x:%04x:%04x:%04x\n" \
        $(hex2dec $1) \
        $(hex2dec $2) \
        $(hex2dec $3) \
        $(hex2dec $4) \
        $(hex2dec $5) \
        $(hex2dec $6) \
        $(hex2dec $7) \
        $(hex2dec $8)
}

function check_cidr_range() {
	
	local start_range=$1
	local end_range=$2
	local user_ip=$3

	#echo "start_range $start_range"
	#echo "end_range $end_range"
	#echo "user_ip $user_ip"

	# If IP is within Start IP Range and End IP Range, Return 1 as Found Flag
	if [[ $user_ip -ge $start_range && $user_ip -le $end_range ]]
	then
		echo 1
	else
		echo 0
	fi
}

get_firewall_list_sources () {

	firewall-cmd --permanent --zone=public --list-sources > firewall.tmp
	cidr_ip_array=()
	for source_scan in $(cat firewall.tmp)
	do
		cidr_ip_array+=("$source_scan")
	done
}

get_nodes_ip () {


	echo "Scan firewall source list that allow accessing this node. Please add nodes IP"

	user_ip_array=()
	
	while read -r -p "Add Node IP (Press Enter to complete)   : " ipAddress
        do

		if [ ! -z "$ipAddress" ] 
		then
			user_ip_array+=("$ipAddress")
		else
			if [ ${#user_ip_array[@]} -gt 0 ]
			then
				echo -e "\n"
				break
			else
				echo -e "\nPlease add at least one IP."
			fi
		fi

	done

}

scan_white_ip () {

	get_firewall_list_sources	

	local start_ip_range=""
	local end_ip_range=""
	local ip=""
	
	local whiteIP=1

	for user_ip in "${user_ip_array[@]}"
	do

		for subnet in "${cidr_ip_array[@]}"
		do
			ipv4Flag=0
			ipv6Flag=0

			IFS=/ read cidr_ip prefix_length <<< "${subnet}"

			user_ip_flag=$(ip_format $user_ip)
			cidr_ip_flag=$(ip_format $cidr_ip)

			if [[ "$user_ip_flag" == "$cidr_ip_flag" ]]
			then
				if [[ "$user_ip_flag" == "ipv4" ]]
				then
			  	       fullBit=$(getFullBit 32)
					
					maskTmp=$[$fullBit << (32 - $prefix_length)]  # IPv4 maskTmp is in Dec
					maskBit=$[$fullBit & $maskTmp]
					
					cidr_ip_dec=$(getV4Address $cidr_ip)

					start_range_ipv4=$(($cidr_ip_dec & $maskBit))        
					end_range_ipv4=$(($cidr_ip_dec | ~$maskBit & $fullBit))  
					
					user_ip_dec=$(getV4Address $user_ip)
		
					ipv4Flag=$(check_cidr_range $start_range_ipv4 $end_range_ipv4 $user_ip_dec)
					if [[ $ipv4Flag -gt 0 ]]
					then
						break
					fi

				else
					if [[ "$user_ip_flag" == "ipv6" ]]
					then
						start_range_ipv6=$(getV6Address $cidr_ip $prefix_length)  # This is in Decimal Format
						
						fullBit=$(getFullBit 128)                      
						maskTmp=$[$fullBit << (128 - $prefix_length)]  # IPv6 maskTmp is total host in CIDR 
						
						end_range_ipv6=`expr $start_range_ipv6 + $maskTmp`
						
						ip=$(getV6Address $user_ip)
						
						#echo "User IP $user_ip [$ip]"
					
						ipv6Flag=$(check_cidr_range $start_range_ipv6 $end_range_ipv6 $ip)
						
						if [[ $ipv6Flag -gt 0 ]]
						then
							break
						fi
					fi
				fi

			fi

		done

		if [[ $ipv4Flag -eq 1 || $ipv6Flag -eq 1 ]]
		then
			echo "[INFO] Node $user_ip added to firewall source list check   --- ok"
		else
			whiteIP=0
			echo "[ERROR] Node $user_ip is not added to firewall source list    --- failed"
		fi		
	done
	
	return $whiteIP
}

check_port_range () {

	IFS="/tcp"
	read -ra mercury_port <<< "$1"
	unset IFS
	for port_scan in $(cat firewall.tmp|sed 's\/tcp\\g'|sed 's\/udp\\g')
	do
		if [[ $port_scan == *"-"* ]] ; then
			IFS="-"
			read -ra port_range <<< "$port_scan"
			start_port=${port_range[0]} 
			end_port=${port_range[1]}
			if [[ ${mercury_port[0]} -ge ${start_port} && ${mercury_port[0]} -le ${end_port} ]] ; then
				return 1
				break
			fi	
			unset IFS
		else
			start_port=$port_scan
			if [[ ${mercury_port[0]} -eq ${start_port} ]] ; then
				return 1
				break
			fi	
		fi
	done
	return 0
}

check_port(){
   check_port_range $1
   if [ $? -eq 0 ]
   then
     echo "[ERROR] Port $1 is not added to firewall   --- failed"
   else
     echo "[INFO] Port $1 check   --- ok"
   fi
}

check_firewall() {

	firewall_cnt=`systemctl status firewalld.service|grep "Active: active"|wc -l`

	if [ $firewall_cnt -eq 1 ]
	then

		> firewall.tmp

		declare -a port_array
		port_array=(111/tcp 2049/tcp 20048/tcp)                   # NFS Port 
		port_array+=(30029/tcp 30031/tcp)                         # Analytics Port
		port_array+=(31443/tcp)                                   # Dashboard Port
		port_array+=(30019/tcp)                                   # Discovery Port
		port_array+=(30017/tcp)                                   # Inventory Port
		port_array+=(30026/tcp)                                   # Nifi Port 
		port_array+=(30033/tcp)                                   # Threshold Port

		get_nodes_ip 

		scan_white_ip

		if [ $? -eq 1 ]  # If is in white IP List,no need further scan specifc Port
		then
			return
		fi

		firewall-cmd --zone=public --list-ports |grep tcp > firewall.tmp
	
		suggest_range_port_cnt=`grep 30000-32767/tcp firewall.tmp | wc -l`

		if [ $suggest_range_port_cnt -eq 0 ]
		then	
			# If port range 3000-32767 is not set, futher scan if individual port is set
			for key in "${!port_array[@]}" 
			do
				port="${port_array[$key]}"
				cnt=`grep -w $port firewall.tmp|wc -l`
			
				if [ $cnt -eq 0 ]
				then
					check_port $port
				else
					echo "[INFO] Port $port check   --- ok"
				fi
			done
			rm firewall.tmp
		 else
			echo "[INFO] Port range 30000-32767/tcp check   --- ok"
			check_port 111/tcp
			check_port 2049/tcp
			check_port 20048/tcp
		 fi
	fi
}

check_unzip() {

	unzip_flag=`which unzip |grep unzip|wc -l|awk '{print $1}'`
	if [ $unzip_flag -lt 1 ]
	then
		echo "[ERROR] unzip is not installed   --- failed"
	else
		echo "[INFO] unzip installed check   --- ok"
	fi

}

check_bc() {

	bc_flag=`which bc |grep bc|wc -l|awk '{print $1}'`
	if [ $bc_flag -lt 1 ]
	then
		echo "[ERROR] bc is not installed   --- failed"
	else
		echo "[INFO] bc installed check   --- ok"
	fi

}


check_inotify_limit() {
	max_user_watches=`sysctl fs.inotify.max_user_watches|awk '{print $3}'`
	max_user_instances=`sysctl fs.inotify.max_user_watches|awk '{print $3}'`

	if [ $max_user_watches -lt 64000 ]
	then
		echo "[ERROR] Inotify max user watches is less than 64000   --- failed"
		echo "	Run below command to set inotify max user watches :"
		echo ""
		echo "	echo fs.inotify.max_user_watches=64000 | tee -a /etc/sysctl.conf"
		echo "	sysctl -p"
		echo ""
	else
		echo "[INFO] Inotify max user watches check   --- ok"
	fi


	if [ $max_user_instances -lt 64000 ]
	then
		echo "[ERROR] Inotify max user instances is less than 64000   --- failed"
		echo "	Run below command to set inotify max user instances :"
		echo ""
		echo "	echo fs.inotify.max_user_instances=64000 | tee -a /etc/sysctl.conf"
		echo "	sysctl -p"
		echo ""
	else
		echo "[INFO] Inotify max user instance check   --- ok"
	fi


}

#-----------------------------------------
# Main
#-----------------------------------------
echo -e "\nStart Prerequisite Scan ....\n"
check_firewall
check_nfs
check_unzip
check_bc
check_inotify_limit
echo -e "\nPrerequisite Scan Completed.\n"