#!/bin/bash

PRODUCT_NAME=tncp
scriptName=$(basename "$0")
scriptDir="$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd)"

for parameter in "$@"
do
    if [[ "$parameter" == "--"*"="* ]]; then
        eval "${parameter/--/}"
    fi
done

if [[ -z "$storage" ]]; then
   storage='nfs'
fi

if [[ -z "$repo" ]]; then
   repo='docker.io/persistentsystems'
fi

if [[ -z "$offline" ]]; then
   offline='false'
fi

if [[ $offline = "true" ]]; then
  find $scriptDir/services -name \*.yaml -exec sed -i "s/@sha.*//" {} \;
fi

export storage

echo Creating namespace...
kubectl create namespace $PRODUCT_NAME >> install.log 2>&1
kubectl config set-context --current --namespace=$PRODUCT_NAME >> install.log 2>&1

if [[ -x "$(command -v oc)"  ]]; then
    oc project $PRODUCT_NAME >> install.log 2>&1
    oc adm policy add-scc-to-user anyuid -z  default >> install.log 2>&1
    oc adm policy add-scc-to-user privileged -z  default >> install.log 2>&1
    oc adm policy add-scc-to-user hostaccess -z default >> install.log 2>&1
    oc adm policy add-scc-to-user hostmount-anyuid -z default >> install.log 2>&1
fi

if [[ $storage = "nfs" ]]; then
  echo Installing nfs...
  nodes=`kubectl get nodes --selector='!node-role.kubernetes.io/master' --output=jsonpath={.items[*].metadata.name}`
  capacity=""
  for node in $nodes
  do
    capacity="$capacity$(kubectl get node $node -o jsonpath='{.status.capacity.ephemeral-storage}')=$node "
  done

  NFS_NODE=`echo $capacity | sort | cut -d' ' -f1 | cut -d'=' -f2`
  kubectl label node $NFS_NODE storage=nfs >> install.log 2>&1
  kubectl apply -f $scriptDir/services/nfs/nfs.yaml >> install.log 2>&1

  NFS=""
  while [ -z "$NFS" ]
  do
    NFS=`kubectl get pods --selector=service=nfs | grep nfs | grep Running | cut -d' ' -f1`
  done
fi

if [[ $storage = "ceph" ]]; then
  echo Installing ceph...
  kubectl apply -f $scriptDir/services/ceph/common.yaml >> install.log 2>&1
  kubectl apply -f $scriptDir/services/ceph/operator.yaml >> install.log 2>&1

  CEPH_OPERATOR=""
  while [ -z "$CEPH_OPERATOR" ]
  do
    sleep 10
    CEPH_OPERATOR=`kubectl get pods -n rook-ceph --selector=app=rook-ceph-operator 2>/dev/null | grep operator | grep Running | cut -d' ' -f1`
  done

  kubectl apply -f $scriptDir/services/ceph/cluster.yaml >> install.log 2>&1
  kubectl apply -f $scriptDir/services/ceph/toolbox.yaml >> install.log 2>&1
  kubectl apply -f $scriptDir/services/ceph/storageclass-filesystem.yaml >> install.log 2>&1

  CEPH_TOOLS=""
  while [ -z "$CEPH_TOOLS" ]
  do
    sleep 10
    CEPH_TOOLS=`kubectl get pods -n rook-ceph --selector=app=rook-ceph-tools 2>/dev/null | grep tools | grep Running | cut -d' ' -f1`
  done

  CEPH_STATUS=""
  while [ "$CEPH_STATUS" != "1" ]
  do
    sleep 10
    CEPH_STATUS=`kubectl -n rook-ceph exec -it $(kubectl -n rook-ceph get pod -l "app=rook-ceph-tools" -o jsonpath='{.items[0].metadata.name}') -- ceph status 2>/dev/null | grep "mds: rook-ceph-filesystem" | wc -l`
  done
fi

OPERATOR=`kubectl get statefulset | grep operator | cut -d' ' -f1`
echo Installing operator...
sed < $scriptDir/services/operator/operator.yaml -e "s|\${storage}|${storage}|g" | sed "s|\${repo}|${repo}|g" | sed "s|REPLACE_NAMESPACE|${PRODUCT_NAME}|g" | tee | kubectl apply -n "${PRODUCT_NAME}" -f - >> install.log 2>&1

OPERATOR=""
while [ -z "$OPERATOR" ]
do
  sleep 10
  OPERATOR=`kubectl get pods --selector=service=operator | grep operator | grep Running | cut -d' ' -f1` >> install.log 2>&1
done

if [[ $offline = "true" ]]; then
  kubectl exec operator-0 -- bash -c "find resources -name \*.yaml -exec sed -i "s/@sha.*//" {} \;"
fi

if [[ -z "$cr" || "$cr" == "true" ]]; then
  echo "Creating default TNCP instance..."

cat << EOF | kubectl apply -f - >> install.log 2>&1
apiVersion: tncp.ibm.com/v1
kind: TNCP
metadata:
  name: tncp
  namespace: ${PRODUCT_NAME}
spec:
  license:
    accept: true
  storageClassName: ${storage}
EOF

elif [[ -f "$cr" ]]; then
  echo "Creating TNCP instance from file $cr..."
  kubectl apply -f $cr
elif [[ "$cr" != "false" ]]; then
  echo "TNCP CR $cr yaml file is missing..."
fi

echo Creating services with nodePort...

cat << EOF | kubectl apply -f - >> install.log 2>&1
apiVersion: v1
kind: Service
metadata:
  name: dashboard-external
spec:
  selector:
    service: dashboard
  ports:
  - protocol: "TCP"
    name: "http"
    port: 31080
    targetPort: 31080
    nodePort: 31080
  - protocol: "TCP"
    name: "https"
    port: 31443
    targetPort: 31443
    nodePort: 31443
  type: NodePort
---
apiVersion: v1
kind: Service
metadata:
  name: nifi-external
spec:
  selector:
    service: nifi
  ports:
  - protocol: "TCP"
    name: "http"
    port: 30026
    targetPort: 30026
    nodePort: 30026
  type: NodePort
EOF

echo Install complete.
